#!/usr/bin/env python
# -*- coding: utf-8 -*-

# # 2A.ml - Interprétabilité et corrélations des variables
# 
# Plus un modèle de machine learning contient de coefficients, moins sa décision peut être interprétée. Comment contourner cet obstacle et comprendre ce que le modèle a appris ? Notion de [feature importance](http://scikit-learn.org/stable/auto_examples/ensemble/plot_forest_importances.html).

# In[1]:


from jyquickhelper import add_notebook_menu
add_notebook_menu()


# In[2]:


# Répare une incompatibilité entre scipy 1.0 et statsmodels 0.8.
from pymyinstall.fix import fix_scipy10_for_statsmodels08
fix_scipy10_for_statsmodels08()


# ## Modèles linéaires
# 
# Les modèles linéaires sont les modèles les plus simples à interpréter. A performance équivalente, il faut toujours choisir le modèle le plus simple. Le module [scikit-learn](http://scikit-learn.org/) ne propose pas les outils standards d'analyse des modèles linéaires (test de nullité, valeur propre). Il faut choisir [statsmodels](http://statsmodels.sourceforge.net/) pour obtenir ces informations.

# In[3]:


import numpy
import statsmodels.api as smapi
nsample = 100
x = numpy.linspace(0, 10, 100)
X = numpy.column_stack((x, x**2 - x))
beta = numpy.array([1, 0.1, 10])
e = numpy.random.normal(size=nsample)
X = smapi.add_constant(X)
y = X @ beta + e


# In[4]:


model = smapi.OLS(y, X)
results = model.fit()
results.summary()


# ## Arbres (tree)

# ### Lectures
# 
# * [treeinterpreter](https://github.com/andosa/treeinterpreter)
# * [Making Tree Ensembles Interpretable](https://arxiv.org/pdf/1606.05390v1.pdf) : l'article propose de simplifier une random forest en approximant sa sortie par une somme pondérée d'arbre plus simples.
# * [Understanding variable importances in forests of randomized trees](http://papers.nips.cc/paper/4928-understanding-variable-importances-in-forests-of-randomized-trees.pdf) : cet article explique plus formellement le calcul des termes ``feature_importances_`` calculés par scikit-learn pour chaque arbre et forêts d'arbres (voir aussi [Random Forests, Leo Breiman and Adele Cutler](http://www.stat.berkeley.edu/~breiman/RandomForests/cc_home.htm))

# ### Module treeinterpreter

# In[5]:


from sklearn.datasets import load_iris
iris = load_iris()
X = iris.data
Y = iris.target


# In[6]:


from sklearn.tree import DecisionTreeClassifier
clf2 = DecisionTreeClassifier(max_depth=3)
clf2.fit(X, Y)
Yp2 = clf2.predict(X)


# In[7]:


from sklearn.tree import export_graphviz
export_graphviz(clf2, out_file="arbre.dot")


# In[8]:


import os
cwd = os.getcwd()
from pyquickhelper.helpgen import find_graphviz_dot
dot = find_graphviz_dot()
os.system ("\"{1}\" -Tpng {0}\\arbre.dot -o {0}\\arbre.png".format(cwd, dot))


# In[9]:


from IPython.display import Image
Image("arbre.png")


# In[10]:


from treeinterpreter import treeinterpreter
pred, bias, contrib = treeinterpreter.predict(clf2, X[106:107,:])


# In[11]:


X[106:107,:]


# In[12]:


pred


# In[13]:


bias


# In[14]:


contrib


# ``pred`` est identique à ce que retourne la méthode ``predict`` de scikit-learn. ``bias`` est la proportion de chaque classe. ``contrib`` est la somme des contributions de chaque variable à chaque classe. On note $X=(x_1, ..., x_n)$ une observation.
# 
# $$P(X \in classe(i)) = \sum_i contrib(x_k,i)$$
# 
# Le [code](https://github.com/andosa/treeinterpreter/blob/master/treeinterpreter/treeinterpreter.py) est assez facile à lire et permet de comprendre ce que vaut la fonction $contrib$.

# ### Exercice 1 : décrire la fonction contrib
# 
# La lecture de [Understanding variable importances
# in forests of randomized trees](http://papers.nips.cc/paper/4928-understanding-variable-importances-in-forests-of-randomized-trees.pdf) devrait vous y aider.

# In[15]:


clf2.feature_importances_


# ### Exercice 2 : implémenter l'algorithme
# 
# Décrit dans [Making Tree Ensembles Interpretable](https://arxiv.org/pdf/1606.05390v1.pdf)

# In[16]:





# ## Interprétation et corrélation

# ### Modèles linéaires
# 
# Les modèles linéaires n'aiment pas les variables corrélées. Dans l'exemple qui suit, les variables $X_2, X_3$ sont identiques. La régression ne peut retrouver les coefficients du modèle initial (2 et 8).

# In[17]:


import numpy
import statsmodels.api as smapi
nsample = 100
x = numpy.linspace(0, 10, 100)
X = numpy.column_stack((x, (x-5)**2, (x-5)**2))  # ajout de la même variable
beta = numpy.array([1, 0.1, 2, 8])
e = numpy.random.normal(size=nsample)
X = smapi.add_constant(X)
y = X @ beta + e


# In[18]:


import pandas
pandas.DataFrame(numpy.corrcoef(X.T))


# In[19]:


model = smapi.OLS(y, X)
results = model.fit()
results.summary()


# ### Arbre / tree
# 
# Les arbres de décision n'aiment pas plus les variables corrélées.

# In[20]:


from sklearn.datasets import load_iris
iris = load_iris()
X = iris.data[:,:2]
Y = iris.target


# In[21]:


from sklearn.tree import DecisionTreeClassifier
clf1 = DecisionTreeClassifier(max_depth=3)
clf1.fit(X, Y)


# In[22]:


clf1.feature_importances_


# On recopie la variables $X_1$.

# In[23]:


import numpy
X2 = numpy.hstack([X, numpy.ones((X.shape[0], 1))])
X2[:,2] = X2[:,0]


# In[24]:


clf2 = DecisionTreeClassifier(max_depth=3)
clf2.fit(X2, Y)


# In[25]:


clf2.feature_importances_


# On voit que l'importance de la variable 1 est diluée sur deux variables.

# ### Exercice 3 : variables corrélées pour un arbre de décision
# 
# Un arbre de décision est composé d'un ensemble de fonctions de seuil. Si $X_i > s_i$ alors il faut suivre cette branche, sinon, telle autre. Les arbres de décision ne sont pas sensibles aux problèmes d'échelle de variables. Si deux variables sont corrélées $cor(X_1, X_2)= 1$, l'arbre subit les mêmes problèmes qu'un modèle linéaire. Dans le cas linéaire, il suffit de changer l'échelle $(X_1, \ln X_2)$ pour éviter ce problème. 
# 
# * Pourquoi cette transformation ne change rien pour un arbre de décision ?
# * Quelle corrélation il faudrait calculer pour repérer les variables identiques selon le point de vue d'un arbre de décision ?

# In[26]:





# In[27]: